from __future__ import annotations

import hashlib
import inspect
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from typing import Any

DEFAULT_DATE_FORMAT = "%Y.%m.%d"


def mkdir_p(path: Path | str) -> None:
    Path(path).mkdir(parents=True, exist_ok=True)


def register_in_collection(
    collection: dict[str, Any], collection_name: str, name: str, obj: Any
) -> None:
    if name in collection:
        msg = f"{collection_name} {name} already registered."
        raise ValueError(msg)
    collection[name] = obj


def unregister_from_collection(
    collection: dict[str, Any], collection_name: str, name: str
) -> None:
    if name not in collection:
        msg = f"{collection_name} {name} not registered."
        raise ValueError(msg)
    collection[name].pop()


##############################
####### Hash functions #######
##############################


def string_to_short_hash(string: str) -> str:
    """Convert a string to a short hash."""
    return hashlib.sha1(string.encode()).hexdigest()[:8]


def get_file_hash(file: Path) -> str:
    """Reads the config file and returns a shortened hash."""
    with file.open("rb") as f:
        # requires Python 3.11+
        return hashlib.file_digest(f, "sha256").hexdigest()[:8]


def formatted_today() -> str:
    """Return the current date in the format YYYY.MM.DD"""
    return datetime.now().strftime(DEFAULT_DATE_FORMAT)


def calculate_function_hash(func: Callable[..., Any], *args, **kwargs) -> str:  # type: ignore[no-untyped-def]
    """Calculate the hash of a function based on its source code and parameters.
    This is useful for caching or identifying unique function calls."""
    # encode parameter values to hash
    arg_hash = hashlib.sha256(str(args).encode() + str(kwargs).encode()).hexdigest()
    # encode function source code to hash
    func_hash = hashlib.sha256(inspect.getsource(func).encode()).hexdigest()
    # combine both hashes
    return hashlib.sha256(arg_hash.encode() + func_hash.encode()).hexdigest()[:8]


def generate_save_path(base_path: Path, workflow_name: str, config_path: Path) -> Path:
    """
    Creates a save path for the workflow and returns the generated path.

    @param base_path: Base path for the save location.
    @param workflow_name: Name of the workflow.
    @param config_path: Path to the configuration file.

    returns: Path to the save location.
    """
    today = formatted_today()
    config_hash = get_file_hash(config_path)
    return Path(f"{base_path}/{workflow_name}/{today}/{config_hash}/").resolve()


def merge_dicts(
    dicts: list[dict[str, Any]],
    names: list[str] | None = None,
) -> dict[str, Any]:
    """
    Merge multiple dictionaries into one.
    """
    if names:
        if len(dicts) != len(names):
            msg = "Number of dictionaries and names must match."
            raise ValueError(msg)
        return dict(zip(names, dicts, strict=False))

    merged = {}
    for d in dicts:
        merged.update(d)
    return merged



from __future__ import annotations

from ._import import (
    class_from_type_string,
    instance_from_type_string,
    is_class,
    is_valid_import,
)
from ._version import find_fast_hep_packages, find_hep_packages

__all__ = [
    "class_from_type_string",
    "find_fast_hep_packages",
    "find_hep_packages",
    "instance_from_type_string",
    "is_class",
    "is_valid_import",
]



from __future__ import annotations

import importlib
import inspect
from typing import Any


def is_valid_import(module_path: str, aliases: dict[str, str]) -> bool:
    """Check if a module can be imported."""
    value = aliases.get(module_path, module_path)
    module_path, class_name = value.rsplit(".", 1)
    try:
        # Import the module
        mod = importlib.import_module(module_path)
        # this must be a class
        class_ = getattr(mod, class_name)
        return is_class(class_)
    except ImportError as _:
        return False


def class_from_type_string(type_string: str) -> Any:
    """Get a class from a type string, e.g. 'module.submodule.Class'."""
    module_path, class_name = type_string.rsplit(".", 1)
    mod = importlib.import_module(module_path)
    return getattr(mod, class_name)


def instance_from_type_string(type_string: str, *args, **kwargs) -> Any:  # type: ignore[no-untyped-def]
    """Create an instance from a type string, e.g. 'module.submodule.Class'."""
    return class_from_type_string(type_string)(*args, **kwargs)


def is_class(obj: Any) -> bool:
    """Check if an object is a class.
    Will unwrap the object to make it work for decorated classes."""
    return inspect.isclass(inspect.unwrap(obj))



from __future__ import annotations

import json
from collections.abc import Sequence
from enum import Enum
from itertools import cycle
from typing import Any

from rich.console import Console
from rich.table import Table
from tabulate import tabulate


class DisplayFormats(str, Enum):
    """Display formats for command output"""

    SIMPLE = "simple"
    PIP = "pip"
    TABLE = "table"
    JSON = "json"
    MARKDOWN = "markdown"
    LATEX = "latex"


COLOURS = [
    "cyan",
    "magenta",
    "green",
    "yellow",
    "white",
]


def display(
    data: list[tuple[Any]],
    title: str | None = None,
    headers: list[str] | None = None,
    display_format: DisplayFormats = DisplayFormats.SIMPLE,
) -> None:
    """Display data in console"""
    match display_format:
        case DisplayFormats.SIMPLE | DisplayFormats.PIP:
            separator = "==" if display_format == DisplayFormats.PIP else ": "
            print_simple(data, title, headers, separator=separator)
        case DisplayFormats.TABLE:
            print_table(data, title, headers)
        case DisplayFormats.MARKDOWN:
            print_table_md(data, title, headers)
        case DisplayFormats.LATEX:
            msg = "LaTeX format is not implemented yet."
            raise NotImplementedError(msg)
        case DisplayFormats.JSON:
            console = Console()
            console.print(json.dumps(data, indent=4))
        # case _:
        #     msg = f"Unknown display format: {format}"
        #     raise ValueError(msg)


def print_simple(
    data: list[tuple[Any]],
    title: str | None = None,
    headers: list[str] | None = None,
    separator: str = ":",
) -> None:
    """Print data in simple format"""
    console = Console()
    if title:
        console.rule(
            f"[bold]{title}[/bold]",
            style="cyan",
        )
        # console.print("=" * 10 + f" [bold]{title}[/bold] " + "=" * 10)
    if headers:
        console.print(
            f"{separator}".join(
                f"[{colour}]<{header}>[/]"
                for colour, header in zip(cycle(COLOURS), headers)
            )
        )
    for row in data:
        console.print(
            f"{separator}".join(
                f"[{colour}]{item}[/]" for colour, item in zip(cycle(COLOURS), row)
            )
        )


def print_table(
    data: list[tuple[Any]],
    title: str | None = None,
    headers: list[str] | None = None,
) -> None:
    """Print data in table format"""
    colours = cycle(
        [
            "cyan",
            "magenta",
            "green",
            "yellow",
            "white",
        ]
    )  # Cycle through colors for each row
    # table = packagesList
    if not headers:
        headers = [f"Column {i + 1}" for i in range(len(data[0]))]
    if not title:
        title = "Data Table"
    table = Table(title=title)
    for header in headers:
        table.add_column(header, justify="left", style=next(colours))
    for row in data:
        table.add_row(*[str(item) for item in row])
    console = Console()
    console.print(table)


def print_table_md(
    data: Sequence[Sequence[Any]],
    title: str | None = None,
    headers: Sequence[str] | None = None,
) -> None:
    """Print data in table format"""
    console = Console()
    if title:
        console.rule(
            f"[bold]{title}[/bold]",
            style="cyan",
        )
    if not headers:
        headers = [f"Column {i + 1}" for i in range(len(data[0]))]
    console.print(
        tabulate(
            data,
            headers=headers,
            tablefmt="github",
            colalign=("left", "right"),
        )
    )



"""
Copyright (c) 2025 FAST-HEP. All rights reserved.

fasthep-toolbench: Package for cross-package utility functions
"""

from __future__ import annotations

from ._version import version as __version__

__all__ = ["__version__"]



"""Functions for finding FAST-HEP software"""

from __future__ import annotations

from collections.abc import Callable, Iterator
from importlib.metadata import distributions

from rich.progress import Progress, SpinnerColumn, TextColumn


def __find_package_versions(
    filter_function: Callable[[str], bool],
) -> Iterator[tuple[str, str]]:
    """
    Find the versions of a list of packages
    """
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        task = progress.add_task("Finding packages...", total=0)

        for dist in distributions():
            progress.update(task, advance=1)
            if filter_function(dist.metadata["Name"].lower()):
                yield dist.metadata["Name"].lower(), dist.version
        progress.update(task, completed=0)


def _is_fasthep_package(package_name: str) -> bool:
    """
    Check if a package is a FAST-HEP package
    """
    fast_hep_prefixes = ["fasthep-", "fast-", "scikit-validate"]
    return any(package_name.startswith(prefix) for prefix in fast_hep_prefixes)


def _is_hep_package(package_name: str) -> bool:
    """Check if a package is a HEP package (list will always be incomplete)"""
    # from https://scikit-hep.org/
    basics = ["awkward", "hepunits", "ragged", "vector"]
    data_manip = ["coffea", "formulate", "hepconvert", "uproot", "uproot_browser"]
    histogramming = ["boost-histogram", "hist", "histoprint", "uhi"]
    particles = ["decaylanguage", "particle"]
    fitting = ["goofit", "iminuit", "pyhf"]
    interfaces = ["fastjet", "pyhepmc", "pylhe"]
    visualisation = ["mplhep", "vegascope"]
    misc = ["cibuildwheel", "root", "fsspec-xrootd", "pybind11", "scikit-hep"]
    testing = ["scikit-hep-testdata"]
    hep_packages = (
        basics
        + data_manip
        + histogramming
        + particles
        + fitting
        + interfaces
        + visualisation
        + misc
        + testing
    )
    return package_name in hep_packages


def find_fast_hep_packages() -> list[tuple[str, str]]:
    """
    Find all FAST-HEP packages
    """
    return sorted(__find_package_versions(_is_fasthep_package))


def find_hep_packages() -> list[tuple[str, str]]:
    """
    Find all HEP packages
    """
    return sorted(__find_package_versions(_is_hep_package))



"""Functions for download command"""

from __future__ import annotations

import json
from pathlib import Path

import httpx
from rich.progress import Progress, SpinnerColumn, TextColumn

from ..logger import default_logger

logger = default_logger()


def download_from_url(url: str, destination: str, force: bool = False) -> None:
    """Download a file from a URL"""
    dst = Path(destination)
    if dst.exists() and not force:
        msg = f"File {destination} already exists. Use --force to overwrite. Skipping download."
        logger.warning(msg)
        return
    result = httpx.get(url, follow_redirects=True, timeout=60)
    with dst.open("wb") as file_handle:
        file_handle.write(result.content)


def download_from_json(json_input: str, destination: str, force: bool = False) -> None:
    """Download files specified in JSON input file into destination directory.
    JSON input file should be a dictionary with the following structure:
    {   "file1": "url1", "file2": "url2", ... }
    """
    dst = Path(destination)
    with Path(json_input).open(encoding="utf-8") as json_file:
        data = json.load(json_file)
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        task = progress.add_task("Downloading files...", total=len(data))
        for name, url in data.items():
            progress.update(task, advance=1)
            # TODO: this should be a logger
            output_path = dst / name
            download = progress.add_task(
                f"Downloading {output_path} from {url}", total=None
            )
            download_from_url(url, output_path, force)
            progress.update(download, completed=0)
        progress.update(task, completed=0)



"""Standard logger configurations for the FAST-HEP packages"""

from __future__ import annotations

import sys

import loguru


def logger_with_logfile(logfile: str) -> loguru.Logger:
    """
    Create a logger that writes to both the console and a specified log file.

    Args:
        logfile (str): The path to the log file.

    Returns:
        loguru.Logger: Configured logger instance.
    """
    logger = loguru.logger
    logger.remove()  # Remove default logger
    logger.add(sys.stderr, level="INFO")  # Log to console
    logger.add(
        logfile, level="DEBUG", rotation="10 MB", retention="10 days"
    )  # Log to file
    return logger


def default_logger() -> loguru.Logger:
    """
    Create a default logger.

    Returns:
        loguru.Logger: logger instance.
    """
    return loguru.logger
